/*
 * StormTableReader.java
 *
 * Created on November 7, 2004, 12:04 AM
 */

/**
 *
 * @author  Andrew T. Flowers, K0SM
 */



import java.io.*;
import java.util.*;
import org.apache.commons.net.ftp.*;
//import sun.net.ftp.*;

/**
 * This class handles the downloading of the storm table The storm table for 
 * each radar site is contained in WSR-88D Radar Product #37 (Shortrange 
 * composite reflectivity).  Most of this file is the Composite reflectivity 
 * data in radial format.  This is the same data that
 * the NOAA displays on its composite reflectivity radar map, which is 
 * downloaded the the RainScatter map panel.  This class thows most of the file 
 * away, except for the Radar Lat/Lon and the Storm table contained within.  The
 * parse() method enters other information into memory should these other data 
 * be utilized later.  If you are considering modifying this class have a look
 * at the downloaded file with a hex editor to get an idea of what the format 
 * looks like.  The Storm Table appears in plain ASCII text toward the end of 
 * the file.  It extends a JFrame just incase someone wishes to put an 
 * interface on it.
 */  
public class StormTableReader extends javax.swing.JFrame {
    //this is used to refer back to the main program
    private final RainScatter MAIN_PROGRAM =                   //this is used to refer 
            (RainScatter)RainScatter.getFrames()[0];   //to the main program
    
    private final int RETRY_INTERVAL = 20000;  //milliseconds before retrying the FTP 
                                        //connection if it fails
    
    private boolean isConnected = false;  //true if there exists an open FTP
                                         //connection to a server
    private Vector stormTable = new Vector();
    private Vector radars = new Vector();
    private String currentRadar = "";
    private javax.swing.Timer retryTimer;
    private File test = new File("sntest");
    
    /** Creates new form StormTableReader */
    public StormTableReader() {
        initComponents();
        //loadRadarList();
        
            //retry timer activated when FTP connection fails for some reason
            //it will wait RETRY_INTERVAL (milliseconds) before attempting
            //to reestablish the connection.
        retryTimer = new javax.swing.Timer(RETRY_INTERVAL, 
                new java.awt.event.ActionListener() {
                    public void actionPerformed(java.awt.event.ActionEvent evt) {
                        retryTimerActionPerformed(evt);
            }
        });
        retryTimer.setRepeats(false);
    }
    

    private void initComponents() {//GEN-BEGIN:initComponents

        getContentPane().setLayout(null);

        setTitle("Station List");
        addWindowListener(new java.awt.event.WindowAdapter() {
            public void windowClosing(java.awt.event.WindowEvent evt) {
                exitForm(evt);
            }
        });

        pack();
    }//GEN-END:initComponents

    /** This method opens an FTP connection with the NOAA FTP server and
     * downloads the storms tables for every radar in the radars vector.
     * This file is written to disk and parsed to create a storm table.
     */
    public void downloadData() {
        /**This the following ensures that only one active FTP connection to the 
         *server is open at any one time.  If download data is called while a 
         *connection is still active, the this method will display the following
         *message inthe status and return without opening another connection.  
         *This will occur when the automatic update timer activates before the 
         *the last FTP session has finished downloading all of the radar files. 
         *It will also occur if the user presses the "Update Storms" button 
         *before the last FTP session has fishished.
         */
        if (isConnected) {  //already a connection open
            
            MAIN_PROGRAM.setStatusLabel("Still downloading storm table...will " +
                                    "not update now.", java.awt.Color.RED); 
            return;
        }
        
        //No connection exists, so open a new one....
        MAIN_PROGRAM.setStormTableProgressBarIndeterminate(true);
        
        
        stormTable = new java.util.Vector();
        String server = "tgftp.nws.noaa.gov";
        String user = "ANONYMOUS";
        
        // password is set like this to ensure that the FTP server sees each 
        // connection from this program as unique.
        String passwd = MAIN_PROGRAM.VERSION + '@' + MAIN_PROGRAM.myGrid;
        
        MAIN_PROGRAM.setStatusLabel("Opening FTP conneciton with " + 
                                            server + "...", java.awt.Color.BLACK); 
        
        
        boolean loggedIn = true;   //will be changed to false if login fails

        final FTPClient client = new FTPClient();
       // final FtpClient client = new FtpClient();
        try {
            isConnected = true;
            /*
            client.openServer(server);
            client.login(user, passwd);
            client.binary();
             *
             */
            client.connect(server);
            client.login(user, passwd);
            client.setFileType(FTP.BINARY_FILE_TYPE);
            client.enterLocalPassiveMode();
        }
        catch (Exception e) {  // if unable to log in, try again in 20 seconds
            e.printStackTrace();
            loggedIn = false;
            isConnected = false;
            
            
            MAIN_PROGRAM.setStatusLabel("Unable to establish FTP connection..." +
                            "retrying in " + RETRY_INTERVAL/1000 + " seconds.", 
                               java.awt.Color.RED); 
        
            MAIN_PROGRAM.setStormTableProgressBarIndeterminate(false);
            retryTimer.start();
        }
        
       // If a connection is established, commence downloading parsing the file
        if (loggedIn) {
            final SwingWorker worker = new SwingWorker() {
                public Object construct() {
                    Integer n = new Integer(0);

                    for (int i = 0; i < radars.size(); i++) {
                        currentRadar =  ((String)radars.get(i)).toLowerCase();
                        
                        MAIN_PROGRAM.setStatusLabel("Downloading " +
                        "storm table for " + currentRadar + "... (" +
                        (i+1) + " of " + radars.size() + ")", java.awt.Color.BLACK); 
                        

                        //download composite radar file from NOAA FTP server.
                        // the storm table is appended toward the end and will be
                        // parsed out by parse().
                        try {
                            
                            client.changeWorkingDirectory("~/SL.us008001/DF.of/DC.radar/DS.p37cr/SI." + currentRadar);
                            System.out.println(client.getStatus());
                            FileOutputStream fout = new FileOutputStream(test);

                            client.retrieveFile("sn.last", fout);

                            fout.close();

                        }
                        /** this catch block will be called if the FTP connection
                         * is inturrupted during download.  It will start the 
                         * retry timer.
                         */
                        catch (Exception e) {
                            e.printStackTrace();
                            MAIN_PROGRAM.setStatusLabel(" FTP connection " +
                                "interrupted." + "  Retrying in " +
                                 RETRY_INTERVAL/1000 + " seconds.", java.awt.Color.RED); 
                            
                            isConnected = false;
                            retryTimer.start();
                        }
                        
                        //if still connected, parse the file.  If not the 
                        //connection was interrrupted in the middle of the file
                        //transfer.  That would have been handled by the above 
                        //catch block.
                        if (isConnected) {
                            parse();
                        }
                    }
                    return n;   //not used, but required of SwingWorker 
                }
                
                //after thread is finished, back on EventDisptach thread
                public void finished() {
                    /**If all files downloaded successfully, this try block
                    * will excectue w/o exceptions.  It will throw an exception
                    * if the client is unable to close the connection.  This 
                    * will occur when the connection has closed prematurely. In
                    * that case and exception should have been generated and 
                    * handled by the above try/catch blocks.  We need not worry 
                    * about it here.
                    */
                    try {  
                        //client.closeServer();
                        client.logout();
                        client.disconnect();
                        MAIN_PROGRAM.setStatusLabel("Finished downloading Storm Table.",
                                                        java.awt.Color.BLACK); 
                     
                    }
                    catch (Exception e) {  
                        
                    }
                    isConnected = false;
                    MAIN_PROGRAM.setStormTableProgressBarIndeterminate(false);
                    
                    
                    
                    //callback to main program to find visible storms
                    MAIN_PROGRAM.findPaths();
                }
            };
            worker.start();     //start the above thread
        }
    }
    
    /** Called by downloadData to read the radar information and Storm Table
     * from a file
     */
    private void parse(){
        String tableString = "";
        int lengthOfFile = -1;
        int startOfTable = -1;
        DataInputStream din;
        int productCode;
        int dataOfMessage;  //days since 1/1/1970
        int timeOfMessage;  //seconds since midnight GMT
        int bytesInMessage;
        int siteID;
        int productDestinationID;
        int dataBlocksInMessage;
        
        ////product discription block
        
        double latitude;
        double longitude;
        int height;  //in feet
        int mode; // 0 = maintenance; 1 =clear; 2=precip
        int volumeCoveragePattern;
        int sequenceNumber;
        int volumeScanNumber;
        int volumeScanStartDate;
        int volumeScanStartTime;
        int productGenerationDate;
        int productGenerationTime;
        int offsetToSymbologyBlock; //usually 60 bytes
        int offsetToGraphicsBlock;
        int offsetToTabularBlock;
        
        
        try {
            //First get radar-specific info
            RandomAccessFile raf = new RandomAccessFile(test, "rw");
            lengthOfFile = (int)raf.length();
            raf.seek(50);
            latitude = ((double)raf.readInt() / 1000);
            raf.seek(54);
            longitude = ((double)raf.readInt() / 1000);
            raf.seek(58);
            height = raf.readShort();
            raf.seek(60);
            productCode = raf.readShort();
            raf.seek(62);
            mode = raf.readShort();
            raf.seek(64);
            volumeCoveragePattern = raf.readShort();
            raf.seek(66);
            sequenceNumber = raf.readShort();
            raf.seek(68);
            volumeScanNumber = raf.readShort();
            raf.seek(70);
            volumeScanStartDate = raf.readShort();
            raf.seek(72);
            volumeScanStartTime = raf.readShort();
            
            /** now find the Storm Attribute Table. This table is the fourth 
            * block in the file.  Blocks are defined by the word [0xFF 0xFF].
            */
            int blockCounter = 0;
            
            int z = 0;
            while (blockCounter < 4) {
                
                raf.seek(z);
                if (raf.readByte() == (0x00 - 1)) {  //don't ask me why 0xFF isn't working!
                    z++;
                    if(raf.readByte() == (0x00 - 1)) {
                        blockCounter++;
                        if (blockCounter == 4) {  //start of Storm Table
                            startOfTable = ++z;
                        }
                    }
                }
                else { z++;}  //not 0xFF, so go to next byte
            }
            
            raf.seek(startOfTable + 4);
            byte[] ba = new byte[((int)raf.readShort()) -2 ];
            raf.seek(startOfTable);
            raf.readFully(ba);
            
            for (int i = 0; i < ba.length; i++) {
                byte b = ba[i];
                if (b == 0x00) {
                    b = ba[i + 1];
                    if (b == 0x08) {
                        b = ba[i + 2];
                        if (b == 0x00) {
                            b = ba[i + 3];
                            if (b  == 0x4E) {
                                String str = "";
                                for (int j = 0; j < 74; j++) {
                                    str = str + (char) ba[i + 11 + j];
                                }
                                if (!str.substring(0, 3).equals("STM") && !str.substring(0,5).equals("     ")) {
                                    addToTable(str, latitude , longitude);
                                }
                            }
                        }
                    }
                }
            }
            //now close up the filestream since we are done reading
            raf.close();
            java.awt.geom.Point2D.Double p = getLatLon(latitude, longitude, 235, 50);
        }
        catch (EOFException eofe) { //something wrong with the file format
            
            MAIN_PROGRAM.setStatusLabel("Error parsing storm table for " + currentRadar,
                                                        java.awt.Color.RED); 

        }
        catch (Exception e){
            // Some other exception, just ignore.  
             e.printStackTrace();
        }
    }
    
    /**This method is called when something happens to the FTP connection.  The
     * program will wait a specified amount of time defined by RETRY_INTERVAL 
     * before attempting to reconnect.  When the retray timer activates an 
     * ActionEvent is created and handled here.
     */
    private void retryTimerActionPerformed(java.awt.event.ActionEvent evt) {
        downloadData();
    }
    
       /**This method creates a Storm from a line of the NOAA Attribute table
     * and adds it to stormTable.  Locations are determined from the lat and lon
     * of the reporting radar.
     */
    private void addToTable(String str, double radlat, double radlon) {
        String id;  //storm id
        double lat; //storm latitude
        double lon; //storm longitude
        String tvs; //tornadic vortex signature
        String meso; //mesocyclone signature
        int posh;   //percent possiblity of severe hail
        int poh;    // possibility of hail
        double mxsize;  // max size of hail
        int vil;      //vertically integrated liquie
        int dbz;     // max reflectivity
        double ht;   //height of max reflectivity
        double top;  // top of storm
        int dir;     // direction of mvt in degrees
        int spd;     // speed of movement in knots
        long time = System.currentTimeMillis();   // current cpu time 
        
        id = str.substring(3,5);
        int az = new Integer(trimString(str.substring(6,10))).intValue();
        int ran = new Integer(trimString(str.substring(11,14))).intValue();
        java.awt.geom.Point2D.Double p = getLatLon(radlat, radlon, az, ran);
        lat = p.x;
        lon = p.y;
        tvs = str.substring(15,19);
        meso = str.substring(20, 24);
        
        if (trimString(str.substring(29, 39)).equals("UNKNOWN")) {
            posh = -1;  //used to makr unkown values
            poh = -1;
            mxsize = -1.0;
        }
        else {
            posh = new Integer(trimString(str.substring(25, 29))).intValue();
            poh  = new Integer(trimString(str.substring(30, 33 ))).intValue();
            mxsize = new Double(trimString(str.substring(34, 39 ))).doubleValue();
        }
        vil = new Integer(trimString(str.substring(40, 45 ))).intValue();
        dbz = new Integer(trimString(str.substring(46, 49 ))).intValue();
        ht = new Double(trimString(str.substring(49, 54 ))).doubleValue();
        top = new Double(trimString(str.substring(55, 60 ))).doubleValue();
        if (trimString(str.substring(61, 70)).equals("NEW")) {  //new storm
            dir = -1;
            spd = -1;
        }
        else {
            dir =  new Integer(trimString(str.substring(61, 65 ))).intValue();
            dir = ((dir - 180) + 720) %360;  //convert to direction it is moving, not coming from
            spd = new Integer(trimString(str.substring(66, 69 ))).intValue();
        }
        
        //create a Storm from these values
        Storm stm = new Storm(id, currentRadar, lat,lon, tvs, meso,posh, poh, mxsize, vil, dbz, ht, top, dir, spd, time);
        
        if (stm.getDirection() == -1) {  //storm is new
            stormTable.add(stm);
        }
        else {                           //storm is old
            boolean alreadyInTable = false;
            for (int i = 0; i < stormTable.size(); i ++) {
                Storm s = (Storm)stormTable.get(i);
                if (    (s.getID().equals(stm.getID()))  &&  (s.getRadar().equals(stm.getRadar()))      ) {
                    stormTable.setElementAt(s, i);
                    alreadyInTable = true;
                }
            }
            if (!alreadyInTable) {
                stormTable.add(stm);
            }
        }
        //remove storms older than 10 minutes (not really needed right now
        for (int i = 0; i < stormTable.size(); i++) {
            Storm s = (Storm)stormTable.get(i);
            long t = System.currentTimeMillis();
            
            if ( t - s.getTime() > 72000 ){  //delete storms older than 12min
                stormTable.removeElementAt(i);
            }
        }
        
    }
       
    
    /** Gets the latitude and longitude of a point defined in polar coordinates
     *from a radar location */
    private java.awt.geom.Point2D.Double getLatLon(double radlat, double radlon, int az, int rng) {
        double lat;
        double lon;
        double azd;
        double rngd;
        rngd = (double)rng*Math.PI/(180*60);
        azd = Math.toRadians((double)az);
        radlat = Math.toRadians(radlat);
        radlon = -Math.toRadians(radlon);  //correction for the formula below--  W is negative in NOAA files
        java.awt.geom.Point2D.Double p = new java.awt.geom.Point2D.Double();
        
        lat = Math.asin(Math.sin(radlat)*Math.cos(rngd) + Math.cos(radlat)*Math.sin(rngd)*Math.cos(azd));
        if (Math.cos(lat) == 0) {
            lon = radlon;
        }
        else {
            lon = ((radlon - Math.asin(Math.sin(azd)*Math.sin(rngd)/Math.cos(lat)) + Math.PI) %
            (2*Math.PI) ) - Math.PI;
        }
        
        p.x = Math.toDegrees(lat);  
        p.y = Math.toDegrees(lon);
        
        return p;
    }
    
    /** Gets the heading from lat/lon 1 to lat/lon 2, as observed from lat/lon 1.
     * This algorithm assumes spherical earth.
     */
    private double getHeading(double latitude1, double longitude1, double latitude2, double longitude2) {
        double tc1;  //initial course from point 1
        double d = getDistance(latitude1, longitude1, latitude2, longitude2);
        double lat1 = Math.toRadians(latitude1);
        double lon1 = Math.toRadians(longitude1);
        double lat2 = Math.toRadians(latitude2);
        double lon2 = Math.toRadians(longitude2);
        
        if (Math.sin(lon2-lon1) < 0) {
            tc1 = Math.acos((Math.sin(lat2)-Math.sin(lat1)*Math.cos(d))/(Math.sin(d)*Math.cos(lat1)));
        }
        else{
            tc1 = 2*Math.PI-Math.acos((Math.sin(lat2)-Math.sin(lat1)*Math.cos(d))/(Math.sin(d)*Math.cos(lat1)));
        }
        tc1 = Math.toDegrees(tc1);
        return tc1;
        
    }
    
    /* Gets distance between two points in RADIANS.  Algorithm assumes spherical earth */
    private double getDistance(double latitude1, double longitude1, double latitude2, double longitude2) {  //works
        double d;
        double lat1 = Math.toRadians(latitude1);
        double lon1 = Math.toRadians(longitude1);
        double lat2 = Math.toRadians(latitude2);
        double lon2 = Math.toRadians(longitude2);
        
        d=2*Math.asin(Math.sqrt(Math.pow((Math.sin((lat1-lat2)/2)),2) + Math.cos(lat1)*Math.cos(lat2)*Math.pow((Math.sin((lon1-lon2)/2)),2)));
        return d;
    }
    
    
    public void setRadarList(Vector radarList) {
        radars = new java.util.Vector();  //replace old list
        for (int i = 0; i < radarList.size(); i++) {
            radars.add(getRadarURL((String)(radarList.elementAt(i))));
        }
        
    }
    
    
    
    private String getRadarURL(String str) {
        String line = "";
        String radar = "";
        try{
            File radardata = new File("radarlocations.txt");
            RandomAccessFile raf = new RandomAccessFile(radardata, "r");
            //find str in file
            
            boolean found = false;
            while (!found) {
                line = raf.readLine();
                
                java.util.StringTokenizer st = new java.util.StringTokenizer(line,"|");
                String[] data = new String[st.countTokens()];
                int count = st.countTokens();
                for (int i = 0; i < count; i++) {
                    data[i] = st.nextToken();
                }
                if (data[0].equals(str)) {
                    radar = data[1];
                    found = true;
                }
            }
        }
        catch (IOException e) {e.printStackTrace();}
        return radar;
    }
    
    private String trimString(String str) {
        String s = "";
        for (int i = 0; i < str.length(); i++) {
            char c = str.charAt(i);
            if (c != ' '  && c != '<' && c != '>') {
                s = s + c;
            }
        }
        return s;
    }
    

    public Vector getStormTable() {
        return stormTable;
    }
    
    public boolean isConnected() {
        return isConnected;
    }
    /** User clicks the "close" button in upper right. Hide the window. */
    private void exitForm(java.awt.event.WindowEvent evt) {//GEN-FIRST:event_exitForm
       setVisible(false);
    }//GEN-LAST:event_exitForm
    
    
    // Variables declaration - do not modify//GEN-BEGIN:variables
    // End of variables declaration//GEN-END:variables
   

}
